# -*- encoding: utf-8 -*-
"""
hio.core.uxd.uxding Module
"""
import sys
import platform
import os
import stat
import errno
import socket
import shutil
from contextlib import contextmanager


from ... import help
from ... import hioing
from ...base import doing, filing

logger = help.ogler.getLogger()



class Peer(filing.Filer):
    """Class to manage non-blocking io on UXD (unix domain) socket.
    Use instance method .close() to close socket

    Because Unix Domain Sockets are reliable no need for retry tymer.


    Inherited Class Attributes:
        HeadDirPath (str): default abs dir path head such as "/usr/local/var"
        TailDirPath (str): default rel dir path tail when using head
        CleanTailDirPath (str): default rel dir path tail when creating clean
        AltHeadDirPath (str): default alt dir path head such as  "~"
                              as fallback when desired head not permitted.
        AltTailDirPath (str): default alt rel dir path tail as fallback
                              when using alt head.
        AltCleanTailDirPath (str): default alt rel path tail when creating clean
        TempHeadDir (str): default temp abs dir path head such as "/tmp"
        TempPrefix (str): default rel dir path prefix when using temp head
        TempSuffix (str): default rel dir path suffix when using temp head and tail
        Perm (int): explicit default octal perms such as 0o1700
        Mode (str): open mode such as "r+"
        Fext (str): default file extension such as "text" for "fname.text"

    Inherited Attributes:
        name (str): unique path component used in directory or file path name
        base (str): another unique path component inserted before name
        temp (bool): True means use TempHeadDir in /tmp directory
        headDirPath (str): head directory path
        path (str | None):  full directory or file path once created else None
        perm (int):  numeric os permissions for directory and/or file(s)
        filed (bool): True means .path ends in file.
                       False means .path ends in directory
        mode (str): file open mode if filed
        fext (str): file extension if filed
        file (File | None): File instance when filed and created.
        opened (bool): True means directory path, uxd file, and socket are
                created and opened. False otherwise

    Class Attributes:
        Umask (int): octal default umask permissions such as 0o022
        MaxUxdPathSize (int:) max characters in uxd file path
        MaxGramSize (int): max bytes in in datagram for this transport
        BufSize (int): used to set buffer size for transport datagram buffers


    Attributes:
        umask (int): unpermission mask for uxd file, usually octal 0o022
                     .umask is applied after .perm is set if any
        bc (int | None): count of transport buffers of MaxGramSize
        bs (int): buffer size of transport buffers. When .bc then .bs is calculated
            by multiplying, .bs = .bc * .MaxGramSize. When .bc is None then .bs
            is provided value or default .BufSize
        wl (WireLog): instance ref for debug logging of over the wire tx and rx
        ls (socket.socket): local socket of this Peer

    """
    HeadDirPath = "/usr/local/var"  # default in /usr/local/var
    TailDirPath = "hio/uxd"
    CleanTailDirPath = "hio/clean/uxd"
    AltHeadDirPath = "~"  # put in ~ as fallback when desired not permitted
    AltTailDirPath = ".hio/uxd"
    AltCleanTailDirPath = ".hio/clean/uxd"
    TempHeadDir = "/tmp"
    TempPrefix = "hio_uxd_"
    TempSuffix = "_test"
    Perm = stat.S_ISVTX | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR  # 0o1700==960
    Mode = "r+"
    Fext = "uxd"
    Umask = 0o022  # default
    MaxUxdPathSize = 108
    MaxGramSize = 65535  # 2 ** 16 - 1  default gram size override in subclass
    BufSize = 65535  # 2 ** 16 - 1  default buffersize


    def __init__(self, *, umask=None, bc=None, bs=None, wl=None,
                 reopen=False, clear=True,
                 filed=False, extensioned=True, **kwa):
        """Initialization method for instance.

        Inherited Parameters:
            reopen (bool): True (re)open with this init
                           False not (re)open with this init but later (default)
            clear (bool): True means remove directory upon close when reopening
                          False means do not remove directory upon close when reopening
            filed (bool): True means .path is file path not directory path
                          False means .path is directiory path not file path
            extensioned (bool): When not filed:
                                True means ensure .path ends with fext
                                False means do not ensure .path ends with fext
            See Filing.Filer for other inherited paramters


        Parameters:
            umask (int): unpermission mask for uxd file, usually octal 0o022
            bc (int | None): count of transport buffers of MaxGramSize
            bs (int | None): buffer size of transport buffers. When .bc is provided
                then .bs is calculated by multiplying, .bs = .bc * .MaxGramSize.
                When .bc is not provided, then if .bs is provided use provided
                value else use default .BufSize
            wl (WireLog): instance ref for debug logging of over the wire tx and rx
        """
        self.umask = umask  # only change umask if umask is not None below
        self.bc = bc
        if self.bc:
            self.bs = self.MaxGramSize * self.bc
        else:
            self.bs = bs if bs is not None else self.BufSize


        self.wl = wl
        self.ls = None  # local socket of this Peer, needs to be opened/bound

        super(Peer, self).__init__(reopen=reopen,
                                   clear=clear,
                                   filed=filed,
                                   extensioned=extensioned,
                                   **kwa)



    def actualBufSizes(self):
        """Returns:
            sizes (tuple); duple of the form (int, int) of the (tx, rx)
                actual socket send and receive buffer size

        """
        if not self.ls:
            return (0, 0)

        return (self.ls.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF),
                self.ls.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF))


    def open(self):
        """Opens socket in non blocking mode.

        Returns:
            result (bool): True of opened successfully. False otherwise

        if socket not closed properly, binding socket gets error
            OSError: (48, 'Address already in use')
        """
        if len(self.path) > self.MaxUxdPathSize:
            self.close()
            raise hioing.SizeError(f"UXD path={self.path} too long, > "
                                   f"{self.MaxUxdPathSize}.")

        #create socket ss = server socket
        self.ls = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)

        # make socket address reusable.
        # the SO_REUSEADDR flag tells the kernel to reuse a local socket in
        # TIME_WAIT state, without waiting for its natural timeout to expire.
        self.ls.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        # setup buffers. With uxd sockets only the SO_SNDBUF matters, but set
        # both anyway
        if self.ls.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF) <  self.bs:
            self.ls.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, self.bs)
        if self.ls.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF) < self.bs:
            self.ls.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, self.bs)
        self.ls.setblocking(0) #non blocking socket

        # setup umask
        oldumask = None
        if self.umask is not None: # change umask for the uxd file
            oldumask = os.umask(self.umask) # set new and return old

        #bind to file path
        try:
            self.ls.bind(self.path)
        except OSError as ex:
            if not ex.errno == errno.ENOENT: # No such file or directory
                logger.error("Error opening UxD %s\n %s\n", self.path, ex)
                return False
            try:
                os.makedirs(os.path.dirname(self.path))
            except OSError as ex:
                logger.error("Error making UXD %s\n %s\n", self.path, ex)
                return False
            try:
                self.ls.bind(self.path)
            except OSError as ex:
                logger.error("Error binding UXD %s\n %s\n", self.path, ex)
                return False

        if oldumask is not None: # restore old umask
            os.umask(oldumask)

        self.path = self.ls.getsockname()  # get resolved path after bind
        self.opened = True
        return self.opened


    def reopen(self, clear=True, **kwa):
        """Idempotently open socket by closing first if need be

        Returns:
            result (bool): True of opened successfully. False otherwise

        Inherited Parameters:
            clear (bool): True means remove directory and uxd file upon close
                          False means do not remove directory and uxd file upon close
            See filing.Filer for other inherited parameters
        """
        opened = super(Peer, self).reopen(clear=clear, **kwa)
        if not opened:
            raise hioing.FilerError(f"Failure opening uxd path {self.path}.")

        return self.open()


    def close(self, clear=True):
        """Closes  socket and unlinks UXD file.

        Inherited Parameters:
            clear (bool): True means remove directory/uxd file upon close
                          False means do not remove directory/uxd file upon close
            See filing.Filer for other inherited parameters
        """
        if self.ls:
            self.ls.close() #close socket
            self.ls = None
            self.opened = False

        try:
            result = super(Peer, self).close(clear=clear)  # removes uxd file at end of path only
        except OSError:
            if os.path.exists(self.path):
                raise

        return result  # True means closed successfully


    def receive(self, **kwa):
        """Perform non blocking receive on  socket.

        Returns:
            result (tuple): of form (bytes, str | None) labeled (data, src) where
                data is bytes of data received
                src is str uxd source path or None
                If data empty then returns (b'', None) but always returns duple
        """
        try:
            data, src = self.ls.recvfrom(self.bs)  # data, uxd source path
        except socket.error as ex:
            # ex.args[0] is always ex.errno for better compat
            # the value of a given errno.XXXXX may be different on each os
            # EAGAIN: BSD 35, Linux 11, Windows 11
            # EWOULDBLOCK: BSD 35 Linux 11 Windows 140
            if ex.args[0]  in (errno.EAGAIN, errno.EWOULDBLOCK):
                return (b'', None) #receive has nothing empty string for data
            else:
                logger.error("Error receive on UXD %s\n %s\n", self.path, ex)
                raise #re raise exception ex

        if self.wl:
            self.wl.writeRx(data, who=src)

        return (data, src)


    def send(self, data, dst, **kwa):
        """Perform non blocking send on socket.

        Returns:
            cnt (int): number of bytes actually sent, may be less than len(data).

        Parameters:
           data (bytes): payload to send
           dst (str):  uxd destination path
        """
        try:
            cnt = self.ls.sendto(data, dst)  # count == int number of bytes sent
        except OSError as ex:
            # ex.args[0] == ex.errno for better compat
            # the value of a given errno.XXXXX may be different on each os
            # EAGAIN: BSD 35, Linux 11, Windows 11
            # EWOULDBLOCK: BSD 35 Linux 11 Windows 140
            if (ex.args[0] in (errno.EAGAIN,
                               errno.EWOULDBLOCK,
                               errno.ENOBUFS,
                               errno.ENOMEM)):
                # not enough buffer space to send, do not consume data
                return 0  # try again later with same data

            else:
                logger.error("Error send UXD from %s to %s.\n %s\n", self.path, dst, ex)
                cnt = 0
                raise

        if self.wl:# log over the wire send
            self.wl.writeTx(data[:cnt], who=dst)

        return cnt




@contextmanager
def openPeer(cls=None, name="test", temp=True, reopen=True, clear=True,
             filed=False, extensioned=True, **kwa):
    """
    Wrapper to create and open UXD Peer instances
    When used in with statement block, calls .close() on exit of with block

    Parameters:
        cls (Class): instance of subclass instance
        name (str): unique identifer of peer. Unique path part so can have many
            Peers each at different paths that each use different dirs or files
        temp (bool): True means open in temporary directory, clear on close
                     Otherwise open in persistent directory, do not clear on close
        reopen (bool): True (re)open with this init
                       False not (re)open with this init but later (default)
        clear (bool): True means remove directory upon close when reopening
                      False means do not remove directory upon close when reopening
        filed (bool): True means .path is file path not directory path
                      False means .path is directiory path not file path
        extensioned (bool): When not filed:
                            True means ensure .path ends with fext
                            False means do not ensure .path ends with fext

    See filing.Filer and uxding.Peer for other keyword parameter passthroughs

    Usage:
        with openPeer() as peer0:
            peer0.receive()

        with openPeer(cls=PeerBig) as peer0:
            peer0.receive()

    """
    peer = None
    if cls is None:
        cls = Peer
    try:
        peer = cls(name=name, temp=temp, reopen=reopen, clear=clear,
                   filed=filed, extensioned=extensioned, **kwa)

        yield peer

    finally:
        if peer:
            peer.close(clear=peer.temp or clear)




class PeerDoer(doing.Doer):
    """Basic UXD Peer Doer
    Stub override in sub class

    Because Unix Domain Sockets are reliable no need for retry tymer.

    To test in WingIde must configure Debug I/O to use external console
    See Doer for inherited attributes, properties, and methods.

    See Doer for inherited attributes, properties, and methods.

    Attributes:
       .peer is UXD Peer instance

    """

    def __init__(self, peer, **kwa):
        """
        Initialize instance.

        Parameters:
           peer is UXD Peer instance
        """
        super(PeerDoer, self).__init__(**kwa)
        self.peer = peer


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter
        """
        # inject temp into file resources here if any
        self.peer.reopen(temp=temp)


    def recur(self, tyme):
        """"""
        # service receives and sends


    def exit(self):
        """"""
        self.peer.close(clear=True)




# -*- encoding: utf-8 -*-
"""
hio.core.multidoing Module

Provides multiprocessing support



"""
import os
import sys
import time
import logging
import json
import signal
import multiprocessing as mp

from collections import deque, namedtuple
from dataclasses import dataclass, astuple, asdict, field
import typing


from . import tyming
from .doing import Doist, Doer
from .. import hioing
from .. import help
from ..help import timing, helping
from ..help.helping import RawDom
from ..help import ogling

from ..help.naming import Namer
from ..core.uxd import PeerMemoer

ogler = ogling.initOgler(prefix='hio_mp', name="Boss", level=logging.ERROR)





"""Bossage (namedtuple):
BossDoer info to be injected in CrewDoer as BossDoer enter time
Fields:
   name (str | None): name of Boss for resource management
   path (str | None): UXD cmd memo path used by crew to talk to their boss
"""
Bossage = namedtuple("Bossage", "name path")

"""Loadage (namedtuple):
BossDoer info to be injected into CrewDoer .start() containing both crew doist
parms for Process target kwargs and and CrewDoer parms

Fields:
   name (str): child doist identifier for resources.
   tyme (float): child doist start tyme
   tock (float | None): child doist tock, tyme lag between runs
   real (bool): child doist True means run in real time, tyme is real time
   limit (float | None): child doist tyme limit. None means run forever
   doers (list[Doers]): child doist List of Doers
   temp (bool | None): True means use temporary file resources
   boss (Bossage | None): BossDoer  info for CrewDoer. May be filled at enter time
                      CrewDoer uses to contact BossDoer.
"""
Loadage = namedtuple("Loadage", "name tyme tock real limit doers temp boss")


"""RawDom hidden methods


    Inherited Class Methods:
        _fromdict(cls, d: dict): return dataclass converted from dict d
        _fromjson(cls, s: str|bytes): return dataclass converted from json s
        _fromcbor(cls, s: bytes): return dataclass converted from cbor s
        _frommgpk(cls, s: bytes): return dataclass converted from mgpk s

    Inherited Methods:
        __iter__(self): asdict(self)
        _asdict(self): return self converted to dict
        _asjson(self): return bytes self converted to json
        _ascbor(self): return bytes self converted to cbor
        _asmgpk(self): return bytes self converted to mgpk

"""
@dataclass
class HandDom(RawDom):
    """Configuration dataclass of CrewDoer crew hand info managed by its BossDoer
    boss.

    Attributes:
        proc (typing.Any | None): crew hand subprocess or None
        exiting (bool): True means commanded to exit but may not have exited yet
                        False means not yet commanded to exit

    """
    proc: typing.Any = None  # crew hand subprocess
    exiting: bool = False  # True means commanded to exit but may not have exited yet


@dataclass
class CrewDom(RawDom):
    """Configuration dataclass of crew subprocess Doist parameters for crew doist
    and its CrewDoer to be injected by BossDoer.
    Use this when storing configuration in database or file. Use RawDom
    serialization hidden methods:


    Attributes:
        name (str): child doist identifier for resources.
        tyme (float): child doist start tyme
        tock (float | None): child doist tock, tyme lag between runs
        real (bool): child doist True means run in real time, tyme is real time
        limit (float | None): child doist tyme limit. None means run forever
        doers (list[Doers]): child doist List of Doers
        temp (bool | None): True means use temporary file resources
        boss (Bossage | None): BossDoer  info for CrewDoer. May be filled at enter time
                                CrewDoer uses to contact BossDoer.
    """
    name: str ='child'  # unique identifier of child process and associated resources
    tyme: float = 0.0    # child doist start tyme
    tock: float | None = None  # child doist tyme lag between runs, None means ASAP
    real: bool = True  # child doist True means run in real time, tyme is real time
    limit: float | None = None  # child doist tyme limit. None mean run forever
    doers: list = field(default_factory=list)  # child doist list of doers
    temp: bool | None = False  # use temporary file resources if any
    boss: (Bossage | None) = Bossage(name=None, path=None)  # BossDoer info


@dataclass
class MemoDom(RawDom):
    """Inter Boss Crew Hand structured memo dataclass. Used for control messages
    Between Boss and Crew Doers via their .peer UXD BossMemoer or CrewMemoer.

    Attributes:
        name (str): unique identifier as source of memo
        tag (str): type of memo
        load (dict): type specific payload of memo
    """
    name: str ='hand'  # unique identifier of source
    tag: str = 'REG'    # type of memo
    load: dict = field(default_factory=dict)  # type specific payload


@dataclass
class AddrDom(RawDom):
    """Load Field Value of ACK

    Attributes:
        name (str): unique identifier as source of memo being acked
        tag (str): type of memo being acked
        addr (dict): addr of source of memo being acked
    """
    name: str ='hand'  # unique identifier of source of ack
    tag: str = 'REG'    # type of memo being acked
    addr: str = ''  # addr of source of acked memo



@dataclass
class AckDom(RawDom):
    """Inter Boss Crew Hand structured memo dataclass. Used for control messages
    Between Boss and Crew Doers via their .peer UXD BossMemoer or CrewMemoer.


    Attributes:
        name (str): unique identifier as source of memo
        tag (str): type of memo
        load (AddrDom): info of acked memo
    """
    name: str ='boss'  # unique identifier of source
    tag: str = 'ACK'    # type of memo
    load: AddrDom = field(default_factory=AddrDom)  # instance of AddrDom



class MultiDoerBase(Namer, PeerMemoer, Doer):
    """MultiDoerBase is base class for Doers in multiprocessing. Each subclass
    has support for UXD peer communications via PeerMemoer super class. As well
    as support for local logging given scope of module global ogler at enter time.

    See Namer, PeerMemoer, and Doer for inherited attributes, properties, and methods.

    Inherited Class Attributes:
        See Namer, PeerMemoer and Doer Classes

    Inherited Attributes:  (See Doer and PeerMemoer for all)
        done (bool): completion state:
                     True means completed fully. Otherwise incomplete.
                     Incompletion value may be None or False.
        opts (dict): schedulaer injects options from .opts into its .do generator
                     function as **opts parameter.
        name (str): unique identifier for this instance used to manage local resources
        temp (bool): True means logger or other file resources created by
                        .start() will use temp
                     False otherwise
        reopen (bool): True (re)open with this init
                       False not (re)open with this init but later (default)
        bc (int | None): count of transport buffers of MaxGramSize

    Attributes:
        logger (Logger | None): from module scope ogler created at enter time
                        with local resources.
        graceful (bool): indication to gracefully exit on next recur. Set by
                         .force signal handler. Must add check in .self.recur
                         if self.graceful: sys.exit()
                         True means exit on next recur
                         False otherwise.

    Inherited Properties:
        tyme (float): is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        tymth (closure): function wrapper closure returned by Tymist.tymen()
                        method. When .tymth is called it returns associated
                        Tymist.tyme. Provides injected dependency on Tymist
                        tyme base.
        tock (float): desired time in seconds between runs or until next run,
                 non negative, zero means run asap
        addrByName (dict): mapping between (name, address) pairs, these
            must be one-to-one so that inverse is also one-to-one
        nameByAddr (dict): mapping between (address, name) pairs, these
            must be one-to-one so that inverse is also one-to-one
        countNameAddr (int): count of entries in .addrByName


    Properties:

    Inherited Methods:
        __call__()  makes instance callable as generator function returning generator
        do() generator method that returns generator
        enter() is enter context action method
        recur() recur context action method or generator method
        clean() clean context action method
        exit() exit context method
        close() close context method
        abort() abort context method
        wind()  injects ._tymth dependency from associated Tymist to get its .tyme
        clearAllNameAddr()
        getAddr(name)
        getName(addr)
        addNameAddr(name, addr)
        remNameAddr(name=None, addr=None)
        changeAddrAtName(name=None, addr=None)
        changeNameAtAddr(addr=None, name=None)

    """

    def __init__(self, *, name='base', temp=False, reopen=False, bc=4, **kwa):
        """Initialize instance.

        Inherited Parameters:  (see Doer and PeerMemoer for all)
            name (str): unique identifier for this BossDoer boss to be used
                        to manage local resources
            temp (bool): True means logger or other file resources created by
                            .start() will use temp
                         False otherwise
            reopen (bool): True (re)open with this init
                           False not (re)open with this init but later (default)
            bc (int | None): count of transport buffers of MaxGramSize

        Parameters:


        """
        super(MultiDoerBase, self).__init__(name=name,
                                            temp=temp,
                                            reopen=reopen,
                                            bc=bc,
                                            **kwa)

        self.logger = None  # assign later from ogler in enter time/scope
        self.graceful = False  # indiction by .force signal handler to exit on next recur



    def force(self, signum, frame):  # signal handler for forced but graceful exit
        self.graceful = True  # exit gracefully in recur
        self.logger.debug("Caught signal=%d so graceful forced exit.", signum)


    def enter(self, *, temp=None):
        """Do 'enter' context.
        Set up resources. Comparable to context manager enter.
        Start processes with config from .loads
        Not a generator method.

        Parameters:
            temp (bool | None): True means use temporary file resources if any.
                                None means ignore parameter value. Use self.temp.

        Inject temp or self.temp into file resources here if any
        Doist or DoDoer winds its doers on enter

        """
        self.logger = ogler.getLogger()  # uses ogler in enter scope
        signal.signal(signal.SIGINT, self.force)  # register signal handers
        signal.signal(signal.SIGTERM, self.force) # register signal handers


    def recur(self, tyme):
        """Do 'recur' context.
        Must be overidden in subclass
        """
        if self.graceful:  # signal handler.force caught signal so exit here
            sys.exit()

        self.service()

        return False  # this  is just a default




class BossDoer(MultiDoerBase):
    """BossDoer spawns multiple crew hand subprocesses and injects each with
    a Doist and Doers. The boss Doists runs the BossDoer in the parent process.
    Each crew hand Doist runs a CrewDoer that coordinates with the BossDoer.

    Analogy Boss runs a Crew of Hans. The parent process has a boss Doist which
    runs a BossDoer. Each crew hand is a child process with its own crew doist
    that runs its own CrewDoer

    See MultiDoerBase for all inherited attributes, properties, and methods.

    Inherited Class Attributes:
        See MultiDoerBase Class


    Inherited Attributes:  (See MultiDoerBase Class for all)
        name (str): unique identifier for this instance
                    used to manage local resources
        logger (Logger | None): from module scope ogler created at enter time
                        with local resources.
        graceful (bool): indication to gracefully exit on next recur. Set by
                         .force signal handler. Must add check in .self.recur
                         if self.graceful: sys.exit()
                         True means exit on next recur
                         False otherwise.

    Attributes:
        loads (list[dict]): BossDoer info to be injected into CrewDoer .start()
                            containing both crew doist parmss for Process target
                            kwargs and and CrewDoer parms
                            (see Loadage._asdict() or CrewDom._asdict())
        ctx (mp.context.SpawnContext | None): context under which to spawn processes
        crew (dict): values HandDom instances keyed by name
        crewed (bool): True means all crew members have registered memo interface
                            with this boss.
                       False means not yet


    Inherited Properties:
        See MultiDoerBase Class

    Properties:

    Inherited Methods:
        See MultiDoerBase Class
    """

    def __init__(self, *, name='boss',loads=None, **kwa):
        """Initialize instance.

        Inherited Parameters:  (see Doer and PeerMemoer for all)
            name (str): unique identifier for this instance to be used
                        to manage local resources

        Parameters:
            loads (list[dict]): parameters used to spinup crew hand subprocess
                                .start(). See fields of Loadage and Bossage

        """
        super(BossDoer, self).__init__(name=name, **kwa)
        self.loads = loads if loads is not None else []
        self.ctx = mp.get_context('spawn')
        self.crew = {}  # dict of HandDom instances keyed by crew name
        self.crewed = False  # True means crew successfully registered with boss



    def enter(self, *, temp=None):
        """Do 'enter' context.
        Set up resources. Comparable to context manager enter.
        Start processes with config from .loads
        Not a generator method.

        Parameters:
            temp (bool | None): True means use temporary file resources if any.
                                None means ignore parameter value. Use self.temp.

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter
        """
        super(BossDoer, self).enter(temp=temp)

        self.logger.debug("BossDoer Enter: name=%s size=%d, ppid=%d, pid=%d, "
                          "module=%s, temp=%s, ogler=%s tyme=%f.",
                          self.name, len(self.loads), os.getppid(), os.getpid(),
                          __name__, temp, ogler.name, self.tyme)

        self.reopen(temp=temp)
        self.logger.debug("Boss name=%s path=%s opened=%s.",
                    self.name, self.path, self.opened)
        boss = Bossage(name=self.name, path=self.path)  # use peer at enter

        for load in self.loads:
            name = load["name"]
            doers = load["doers"]
            if not doers or not isinstance(doers[0], CrewDoer):  # first doer must be CrewDoer
                raise hioing.MultiError(f"No crew doer for crew hand doist={name}.")

            doers[0].boss = boss  # assign boss contact info to crew doer
            doers[0].name = name  # make name of CrewDoer same as name of crew doist

            if name not in self.crew:  # ensure unique by name
                proc = self.ctx.Process(name=name,
                                         target=self.start,
                                         kwargs=load)
                dom = HandDom(proc=proc, exiting=False)
                self.crew[proc.name] = dom
                proc.start()
            else:
                raise hioing.MultiError(f"Non-unique crew hand {name=} in loads.")



    def recur(self, tyme):
        """Do 'recur' context."""
        #self.logger.debug("BossDoer Recur: name=%s, pid=%d, ogler=%s, tyme=%f.",
                          #self.name, os.getpid(), ogler.name, tyme)

        if self.graceful:  # signal handler.force caught signal so exit here
            sys.exit()

        self.service()

        if self.crewed:
            if not self.ctx.active_children():
                return True  # childred exited on their own so unforced exit

        # otherwise run forever

        return False  # incomplete recur again


    def exit(self):
        """Do 'exit' (try finally) context."""
        self.logger.debug("BossDoer Exit: name=%s, ppid=%d, pid=%d, module=%s, "
                          "ogler=%s, tyme=%f.", self.name, os.getppid(),
                          os.getpid(), __name__, ogler.name, self.tyme)

        self.close(clear=True)

        self.logger.debug("Boss name=%s path=%s opened=%s.",
                self.name, self.path, self.opened)

        for name, dom in self.crew.items():  # dom is CrewDom instance
            if dom.proc.is_alive():
                dom.proc.terminate()  # force exit with SIGTERM
                self.logger.debug("Boss name=%s pid=%d terminating hand name=%s "
                                  "pid=%d.", self.name, os.getpid(), name,
                                  dom.proc.pid)


    def cease(self):
        """Do 'cease' context."""


    def abort(self, ex):
        """Do 'abort' context."""


    def serviceRxMemos(self):
        """Service all memos in .rxms (greedy) if any

        Override in subclass to handle result(s) and put them somewhere
        """
        while self.rxms:
            memo, src, vid = self._serviceOneRxMemo()
            self.logger.debug("Boss Peer RX: name=%s rx from src=%s memo=%s.",
                                self.name, src, memo)
            memo = json.loads(memo)
            tag = memo['tag']
            if tag == "REG":
                name = memo['name']
                try:
                    self.addNameAddr(name=name, addr=src)
                except hioing.NamerError as ex:
                    self.changeAddrAtName(name=name, addr=src)
                if self.countNameAddr == len(self.crew):
                    self.crewed = True
                    self.logger.debug("Boss name=%s crewed=%s with size=%d at "
                                      "tyme=%f.", self.name, self.crewed,
                                      len(self.crew), self.tyme)

            dst = src
            mack = dict(name=self.name, tag="ACK", load=dict(tag='REG', name=name, addr=src))
            mack = json.dumps(mack,separators=(",", ":"),ensure_ascii=False)
            self.memoit(mack, dst)



    @staticmethod
    def start(*, name='crew', tyme=0.0, tock=None, real=True, limit=None,
               doers=None, temp=None, boss=None):
        """Process target function to spinup and run doist inside crew subprocess
        after it has been started.

        Parameters:
            name (str): unique crew hand name to be used to manage resources
            tyme (float): crew doist initial value of cycle time in seconds
            tock (float | None): crew doist tock time in seconds. None means run Asap
            real (bool): crew doist True means run in real time,
                        Otherwise run faster than real
            limit (float | None): crew doist seconds for max run time of doist.
                                  None means no limit.
            doers (iterable[Doer] | None): crew doist Doer class instances
                                   First entry must be CrewDoer
            temp (bool | None): True means use temp file resources by injection.
                                Otherwise ignore do not inject.
            boss (Bossage | None): boss info. May be filled at enter time
                                  CrewDoer uses to contact BossDoer.


        Doist must be built after process started so local tymth closure is created
        inside subprocess so that when doist winds the tyme for its doers
        the tymth closure is locally sourced.

        When run inside subprocess, this static method provides the outside
        scope for any Doist and doers that reference objects inside the
        __main__ moduleo that calls this method such as ogler.
        The subprocess gets a picked copy of these module globals. Any Doer
        inside then has access to the module globals and can update them as
        neede. In the case of ogler this means changing ogler.level, ogler.temp
        and running ogler.reopen(temp=temp) as appropriate.
        """
        logger = ogler.getLogger()  # uses ogler from subprocess scope

        logger.debug("Crew Start: name=%s, ppid=%d, pid=%s, module=%s, temp=%s, ogler=%s.",
                        name, os.getppid(), os.getpid(), __name__, temp, ogler.name)
        time.sleep(0.01)

        doist = Doist(name=name, tyme=tyme, tock=tock, real=real, limit=limit,
                      doers=doers, temp=temp)
        try:
            doist.do()

        finally:
            logger.debug("Crew End: name=%s, ppid=%d, pid=%s, module=%s, temp=%s, "
                         "ogler=%s, doist.tyme=%f.", name, os.getppid(), os.getpid(),
                         __name__, temp, ogler.name, doist.tyme)



class CrewDoer(MultiDoerBase):
    """CrewDoer runs interface between a given crew hand subprocess and its
    boss process. This must be first doer run by crew hand subprocess doist.

    See MultiDoerBase for all inherited attributes, properties, and methods.

    Inherited Class Attributes:
        See MultiDoerBase Class


    Inherited Attributes:  (See MultiDoerBase Class for all)
        name (str): unique identifier for this instance
                    used to manage local resources
        logger (Logger | None): from module scope ogler created at enter time
                        with local resources.
        graceful (bool): indication to gracefully exit on next recur. Set by
                         .force signal handler. Must add check in .self.recur
                         if self.graceful: sys.exit()
                         True means exit on next recur
                         False otherwise.

    Attributes:
        boss (Bossage | None): contact info for communicating with boss
        registered (bool): True means .path acked registered with boss memoing
                           False not yet registered


    Inherited Properties:
        See MultiDoerBase Class

    Properties:

    Inherited Methods:
        See MultiDoerBase Class

    """

    def __init__(self, *, name='crew', boss=None, **kwa):
        """Initialize instance.

        Inherited Parameters:
            name (str): unique identifier for this instance to be used
                        to manage local resources
            temp (bool): True means logger or other file resources created by

        Parameters:
            boss (Bossage): contact info for BossDoer. assigned by boss at enter


        """
        super(CrewDoer, self).__init__(name=name, **kwa)
        self.boss = boss if boss is not None else Bossage(name=None, path=None)
        self.registered = False  # True means .path acked registered with boss memoing


    def force(self, signum, frame):  # signal handler for forced but graceful exit
        self.graceful = True  # exit gracefully in recur
        self.logger.debug("Doer name=%s caught signal=%d so graceful forced exit.",
                          self.name, signum)


    def enter(self, *, temp=None):
        """Do 'enter' context.
        Set up resources. Comparable to context manager enter.
        Not a generator method.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter
        """
        super(CrewDoer, self).enter(temp=temp)

        self.logger.debug("CrewDoer Enter: hand name=%s pid=%d, temp=%s, ogler=%s, "
                          "tyme=%f.", self.name, os.getpid(), temp, ogler.name,
                          self.tyme)

        self.logger.debug("Hand name=%s Boss: name=%s path=%s.",
                            self.name, self.boss.name, self.boss.path)
        self.addNameAddr(name=self.boss.name, addr=self.boss.path)

        self.reopen(temp=temp)

        self.logger.debug("Hand name=%s path=%s opened=%s.",
                    self.name, self.path, self.opened)

        memo = dict(name=self.name, tag="REG", load={})
        memo = json.dumps(memo, separators=(",", ":"), ensure_ascii=False)
        dst = self.boss.path
        self.memoit(memo, dst)



    def recur(self, tyme):
        """Do 'recur' context."""
        self.logger.debug("CrewDoer Recur: hand name=%s, pid=%d, ogler=%s, tyme=%f.",
                          self.name, os.getpid(), ogler.name, tyme)
        if self.graceful:  # signal handler.force caught signal so exit here
            sys.exit()

        self.service()

        #if self.registered and tyme > self.tock * 3:
            #self.logger.debug("Hand name=%s registered and done at tyme=%f.",
                              #self.name, tyme)
            #return True  # complete

        #if tyme > self.tock * 20:
            #return True

        return False  # incomplete


    def exit(self):
        """Do 'exit' (try finally) context."""
        self.close(clear=True)

        self.logger.debug("CrewDoer Exit: hand name=%s pid=%d, ogler=%s, tyme=%f.",
                          self.name, os.getpid(), ogler.name, self.tyme)
        self.logger.debug("Hand name=%s path=%s opened=%s.",
                    self.name, self.path, self.opened)


    def cease(self):
        """Do 'cease' context."""



    def abort(self, ex):
        """Do 'abort' context."""



    def serviceRxMemos(self):
        """Service all memos in .rxms (greedy) if any

        Override in subclass to handle result(s) and put them somewhere
        """
        while self.rxms:
            memo, src, vid = self._serviceOneRxMemo()
            self.logger.debug("Hand Peer RX: name=%s rx from src=%s memo=%s.",
                                self.name, src, memo)
            memo = json.loads(memo)
            tag = memo['tag']

            if tag == "END":
                name = memo["name"]
                if name == self.boss.name and src == self.boss.path:
                    self.logger.debug("Hand name=%s exiting.", self.name)
                    self.graceful = True  # next recur graceful exit

            elif tag == "ACK":
                name = memo["name"]
                if name == self.boss.name and memo['load']['tag'] == 'REG':
                    self.registered = True
                    self.logger.debug("Hand name=%s registered with boss=%s",
                                        self.name, self.boss.name)





# -*- encoding: utf-8 -*-
"""
hio.help.wiring module

"""

import os
import datetime
import io
import tempfile
import shutil
from contextlib import contextmanager

from .. import hioing
from ..base import doing

@contextmanager
def openWL(cls=None, name="test", temp=True, **kwa):
    """
    Context manager wrapper WireLog instances.
    Defaults to temporary wire logs.
    Context 'with' statements call .close on exit of 'with' block

    Parameters:
        cls is Class instance of subclass instance
        name is str name of wirelog instance for filename so can have multiple wirelogs
             at different paths thar each use different file name
        temp is Boolean, True means open in temporary directory, clear on close
                Otherwise open in persistent directory, do not clear on close

    Usage:

    with openWL(name="bob") as wl:
        wl.writeRx  ....

    with openWL(name="eve", cls=SubclassedWireLog)

    """
    wl = None
    if cls is None:
        cls = WireLog
    try:
        wl = cls(name=name, temp=temp, reopen=True, **kwa)
        yield wl

    finally:
        if wl:
            wl.close()  # if .temp also clears


class WireLog():
    """
    For debugging of non-blocking transports, provides log files or in memory
    buffers for logging 'over the wire' network tx and rx packets as bytes

    Attributes:
        .rxed is Boolean True means log rx
        .txed is Boolean True means log tx
        .samed is Boolean True means log both rx and tx to same file or buffer
        .filed is Boolean True means log to file False means log to memory buffer
        .fmt is io write bytes printf style format string
            Default is b'\n%(dx)b %(who)b:\n%(data)b\n' where:
                who is src or dst for rx tx respectively
                dx is the io direction and will be set to either b'tx' or b'rx' and
                data is the actual io data as bytes
            to write io data without direction who or line feeds use fmt= b'%(data)b'
        .name is str used in file name
        .temp is Boolean True means use /tmp directory
        .prefix is str used as part of path prefix and formating
        .headDirPath is str used as head of path
        .tailDirpath is str used as tail of path
        .altTailDirPath is str used a alternate tail of path
        .dirPath is full directory path
        .rxl is rx log io file or io buffer
        .txl is tx log io file or io buffer
        .opened is Boolean, True means file is opened Otherwise False

    """
    Prefix = "hio"

    HeadDirPath = os.path.join(os.path.sep, 'usr', 'local', 'var')  # default in /usr/local/var

    TailDirPath = "wirelogs"
    AltHeadDirPath = "~"  #  put in ~ as fallback when desired dir not permitted
    TempHeadDir = os.path.join(os.path.sep, "tmp")
    TempPrefix = "test_"
    TempSuffix = "_temp"
    Format = b'\n%(dx)b %(who)b:\n%(data)b\n'  # default format string as bytes


    def __init__(self, rxed=True, txed=True, samed=False, filed=False,
                 fmt=None, name='main', temp=False, prefix=None,
                 headDirPath=None, reopen=False, clear=False):
        """
        Init Loggery factory instance

        Parameters:
            rxed is Boolean True means log rx received bytes
            txed is Boolean True means log tx transmitted bytes
            samed is Boolean True means log both rx and tx to same file or buffer
            filed is Boolean True means log file False means log to memory buffer
            fmt is printf style format str or bytes converted based on .byted
            name is application specific log file name
            temp is Boolean True means use /tmp directory If .filed and clear on close
                    False means use  headDirpath If .filed
            prefix is str to include in path and logging template
            headDirPath is str for custom headDirPath for log file
            reopen is Booblean True means reopen path if anything changed
            clear is Boolean True means clear .dirPath when closing in reopen
        """
        self.rxed = True if rxed else False
        self.txed = True if txed else False
        self.samed = True if samed else False
        self.filed = True if filed else False
        self.fmt = fmt if fmt is not None else self.Format
        if hasattr(self.fmt, "encode"):   # ensure bytes
            self.fmt = self.fmt.encode("utf-8")
        self.name = name if name else 'main'  # for file name
        self.temp = True if temp else False
        self.prefix = prefix if prefix is not None else self.Prefix
        self.headDirPath = headDirPath if headDirPath is not None else self.HeadDirPath
        self.dirPath = None
        self.rxl = None  # receive log iofile or iobuffer
        self.txl = None  # transmit log iofile or iobuffer
        self.opened = False

        if reopen:
            self.reopen(headDirPath=self.headDirPath, clear=clear)


    def reopen(self, rxed=None, txed=None, samed=None, filed=None, fmt=None,
               name=None, temp=None, headDirPath=None, clear=False):
        """
        Use or Create if not preexistent, directory path for file .path
        First closes .path if already opened. If clear is True then also clears
        .path before reopening

        Parameters:
            rxed is optional Boolean If None or unchanged then ignore.
                Otherwise when creating io use .rxed if not provided
            txed is optional Boolean If None or unchanged then ignore.
                Otherwise when creating io use .txed if not provided
            samed is optional Boolean If None or unchanged then ignore.
                Otherwise when creating io use .same if not provided
            filed is optional Boolean If None or unchanged then ignore.
                Otherwise when creating io use .filed if not provided
            fmt is optional bytes printf format If None or unchanged then ignore.
                Otherwise when creating io use .fmt if not provided
            name is optional name
                if None or unchanged then ignore otherwise recreate path
                    When recreating path, If not provided use .name
            temp is optional boolean:
                If None ignore Otherwise
                    Assign to .temp
                    If True then open in temporary directory and clear on close,
                    If False then open persistent directory
            headDirPath is optional str head directory pathname of main database
                if None or unchanged then ignore otherwise recreate path
                   When recreating path, If not provided use default .HeadDirpath
            clear is Boolean True means clear .path when closing
        """
        if self.opened:
            self.close(clear=clear)

        if self.prefix and os.path.isabs(self.prefix):
            raise hioing.FilerError(f"Invalid .prefix={self.prefix} not relative path.")

        # check for changes in creation parameters if need to recreate
        if rxed is not None and rxed == self.rxed:
            rxed = None  # don't need to recreate  because of rxed change
        if txed is not None and txed == self.txed:
            txed = None  # don't need to recreate  because of txed change
        if samed is not None and samed == self.samed:
            samed = None  # don't need to recreate  because of samed change
        if filed is not None and filed == self.filed:
            filed = None  # don't need to recreate  because of filed change
        if fmt is not None and fmt == self.fmt:
            fmt = None
        if name is not None and name == self.name:
            name = None  # don't need to recreate path because of name change
        if temp is not None and temp == self.temp:
            temp = None  # don't need to recreate path because of temp change
        if headDirPath is not None and headDirPath == self.headDirPath:
            headDirPath = None  # don't need to recreate path because of headDirPath change

        if name and os.path.isabs(name):
            raise hioing.FilerError(f"Invalid {name=} not relative path.")

        # always recreates if dirPath is empty or if some creation parameter has changed
        if (not self.dirPath or
            rxed is not None or
            txed is not None or
            samed is not None or
            filed is not None or
            fmt is not None or
            temp is not None or
            headDirPath is not None or
            name is not None):  # need to recreate self.dirPath etc

            if rxed is not None:
                self.rxed = rxed
            if txed is not None:
                self.txed = txed
            if samed is not None:
                self.samed = samed
            if filed is not None:
                self.filed = filed
            if fmt is not None:
                self.fmt = fmt
                if hasattr(self.fmt, "encode"):   # ensure bytes
                    self.fmt = self.fmt.encode("utf-8")

            if self.rxed or self.txed:
                if self.filed:  # use io files
                    if temp is not None:
                        self.temp = True if temp else False
                    if headDirPath is not None:
                        self.headDirpath = headDirPath
                    if name is not None:  # used below for filename
                        self.name = name

                    if self.temp:
                        dirPath = os.path.abspath(
                                            os.path.join(self.TempHeadDir,
                                                         self.prefix,
                                                         self.TailDirPath))
                        if not os.path.exists(dirPath):
                            os.makedirs(dirPath)  # mkdtemp only makes last dir
                        self.dirPath = tempfile.mkdtemp(prefix=self.TempPrefix,
                                                        suffix=self.TempSuffix,
                                                        dir=dirPath)

                    else:
                        self.dirPath = os.path.abspath(
                                            os.path.expanduser(
                                                os.path.join(self.headDirPath,
                                                             self.prefix,
                                                             self.TailDirPath)))

                        if not os.path.exists(self.dirPath):
                            try:
                                os.makedirs(self.dirPath)
                            except OSError as ex:  # can't make dir
                                # use alt=user's directory instead
                                prefix = ".{}".format(self.prefix)  # hide it
                                self.dirPath = os.path.abspath(
                                                os.path.expanduser(
                                                    os.path.join(self.AltHeadDirPath,
                                                                prefix,
                                                                self.TailDirPath)))
                                if not os.path.exists(self.dirPath):
                                    os.makedirs(self.dirPath)
                        else:  # path exists
                            if not os.access(self.dirPath, os.R_OK | os.W_OK):
                                # but can't access it
                                # use alt=user's directory instead
                                prefix = ".{}".format(self.prefix)  # hide it
                                self.dirPath = os.path.abspath(
                                                os.path.expanduser(
                                                    os.path.join(self.AltHeadDirPath,
                                                                 prefix,
                                                                 self.TailDirPath)))
                                if not os.path.exists(self.dirPath):
                                    os.makedirs(self.dirPath)

                    mode = 'wb+'  # bytes io
                    dts = datetime.datetime.now(datetime.timezone.utc).strftime(
                                                         '%Y-%m-%d_%H-%M-%S_%f')
                    if self.samed:
                        fileName = "{}.{}.log".format(self.name, dts)
                        path = os.path.join(self.dirPath, fileName)
                        try:
                            log = io.open(path, mode=mode)
                            if self.rxed:
                                self.rxl = log
                            if self.txed:
                                self.txl = log
                        except IOError:
                            self.rxl = self.txl = None
                            return self.opened

                    else:
                        if self.rxed:
                            fileName = "{}.{}.rx.log".format(self.name, dts)
                            path = os.path.join(self.dirPath, fileName)
                            try:
                                self.rxl = io.open(path, mode=mode)
                            except IOError:
                                self.rxl = None
                                return self.opened

                        if self.txed:
                            fileName = "{}.{}.tx.log".format(self.name, dts)
                            path = os.path.join(self.dirPath, fileName)
                            try:
                                self.txl = io.open(path, mode=mode)
                            except IOError:
                                self.txl = None
                                return self.opened

                else:  # use io memory buffers
                    if self.samed:
                        try:
                            log = io.BytesIO()
                            if self.rxed:
                                self.rxl = log
                            if self.txed:
                                self.txl = log
                        except IOError:
                            self.rxl = self.txl = None
                            return self.opened
                    else:
                        if self.rxed:
                            try:
                                self.rxl = io.BytesIO()
                            except IOError:
                                self.rxl = None
                                return self.opened

                        if self.txed:
                            try:
                                self.txl = io.BytesIO()
                            except IOError:
                                self.txl = None
                                return self.opened

        self.opened = True
        return self.opened


    def flush(self):
        """
        flush files if any and opened.
        A file flush only moves from program buffer to operating system buffer.
        A file fsync moves from operating system buffer to disk.
        """
        if self.filed:
            if not self.rxl.closed:
                self.rxl.flush()
                os.fsync(self.rxl.fileno())
            if not self.txl.closed and not self.samed:
                self.txl.flush()
                os.fsync(self.txl.fileno())


    def close(self, clear=False):
        """
        Close io logs. If clear or self.temp then remove directory at .dirPath
        Parameters:
           clear is boolean, True means clear directory at .dirPath if any
        """
        self.flush()
        self.opened = False
        if self.txl and not self.txl.closed:
            self.txl.close()
        if self.rxl and not self.rxl.closed:
            self.rxl.close()
        if clear or self.temp:
            self.clearDirPath()


    def clearDirPath(self):
        """
        Remove logfile directory at .dirPath
        """
        if self.dirPath and os.path.exists(self.dirPath):
            shutil.rmtree(self.dirPath)


    def readRx(self):
        """
        Returns rx string buffer value if .buffify else None
        """
        if self.rxl and not self.rxl.closed:
            self.rxl.flush()
            self.rxl.seek(0, io.SEEK_SET)
            return (self.rxl.read())
        return None


    def readTx(self):
        """
        Returns tx string buffer value if .buffify else None
        """
        if self.txl and not self.txl.closed:
            self.txl.flush()
            self.txl.seek(0, io.SEEK_SET)
            return (self.txl.read())
        return None


    def writeRx(self, data, who=b''):
        """
        Write bytes data received from source host port address tuple,
        """
        if self.rxed and self.rxl:
            if not isinstance(who, (bytes, bytearray, str)):
                who = str(who)
            if hasattr(who, 'encode'):
                who = who.encode("utf-8")
            if hasattr(data, 'encode'):
                data = data.encode("utf-8")
            self.rxl.write(self.fmt % {b'dx': b'Rx', b'who': who, b'data': data})


    def writeTx(self, data, who=b''):
        """
        Write bytes data transmitted to destination address da,
        """
        if self.txed and self.txl:
            if not isinstance(who, (bytes, bytearray, str)):
                who = str(who)
            if hasattr(who, 'encode'):
                who = who.encode("utf-8")
            if hasattr(data, 'encode'):
                data = data.encode("utf-8")
            self.txl.write(self.fmt % {b'dx': b'Tx', b'who': who, b'data': data})



class WireLogDoer(doing.Doer):
    """
    Basic WireLog Doer

    Inherited Attributes:
        .done is Boolean completion state:
            True means completed
            Otherwise incomplete. Incompletion maybe due to close or abort.

    Attributes:
        .wl is WireLog subclass

    Inherited Properties:
        .tyme is float ._tymist.tyme, relative cycle or artificial time
        .tock is float, desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    Properties:

    Methods:
        .wind  injects ._tymist dependency
        .__call__ makes instance callable
            Appears as generator function that returns generator
        .do is generator method that returns generator
        .enter is enter context action method
        .recur is recur context action method or generator method
        .exit is exit context method
        .close is close context method
        .abort is abort context method

    Hidden:
       ._tymist is Tymist instance reference
       ._tock is hidden attribute for .tock property
    """

    def __init__(self, wl, **kwa):
        """
        Inherited Parameters:
           tymist is Tymist instance
           tock is float seconds initial value of .tock

        Parameters:
           keeper is Keeper instance
        """
        super(WireLogDoer, self).__init__(**kwa)
        self.wl = wl


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter
        """
        # inject temp into file resources here if any
        if not self.wl.opened:
            self.wl.reopen(temp=temp)


    def exit(self):
        """"""
        self.wl.close()



# -*- encoding: utf-8 -*-
"""
hio.help.timing module

"""
import time

from .. import hioing


class TimerError(hioing.HioError):
    """
    Generic Timer Errors
    Usage:
        raise TimerError("error message")
    """

class RetroTimerError(TimerError):
    """
    Error due to real time being retrograded before start time of timer
    Usage:
        raise RetroTimerError("error message")
    """


class Timer(hioing.Mixin):
    """
    Class to manage real elaspsed time using time module.
    Attributes:
        ._start is start tyme in seconds
        ._stop  is stop tyme in seconds

    Properties:
        .duration is float time duration in seconds of timer from ._start to ._stop
        .elaspsed is float time elasped in seconds since ._start
        .remaining is float time remaining in seconds until ._stop
        .expired is boolean, True if expired, False otherwise, i.e. time >= ._stop

    methods:
        .start()  start timer at current time
        .restart() = restart timer at last ._stop so no time lost
    """

    def __init__(self, duration=0.0, start=None, **kwa):
        """
        Initialization method for instance.
        Parameters:
            duration is float duration of timer in seconds (fractional)
            start is float optional start time in seconds allows starting before
               or after current time
        """
        super(Timer, self).__init__(**kwa)  # Mixin for Mult-inheritance MRO
        self._start = float(start) if start is not None else time.time()
        self._stop = self._start + float(duration)  # need for default duration
        self.start(duration=duration, start=start)


    @property
    def duration(self):
        """
        duration property getter,  .duration = ._stop - ._start
        .duration is float duration tyme
        """
        return (self._stop - self._start)


    @property
    def elapsed(self):
        """
        elapsed time property getter,
        Returns elapsed time in seconds (fractional) since ._start.
        """
        return (time.time() - self._start)


    @property
    def remaining(self):
        """
        remaining time property getter,
        Returns remaining time in seconds (fractional) before ._stop.
        """
        return (self._stop - time.time())


    @property
    def expired(self):
        """
        Returns True if timer has expired, False otherwise.
        time.time() >= ._stop,
        """
        return (time.time() >= self._stop)


    def start(self, duration=None, start=None):
        """
        Starts Timer of duration secs at start time start secs.
            If duration not provided then uses current duration
            If start not provided then starts at current time.time()
        """
        # remember current duration when duration not provided
        duration = float(duration) if duration is not None else self.duration
        self._start = float(start) if start is not None else time.time()
        self._stop = self._start + duration
        return self._start


    def restart(self, duration=None):
        """
        Lossless restart of Timer at start = ._stop for duration if provided,
        Otherwise current duration.
        No time lost. Useful to extend Timer so no time lost
        """
        return self.start(duration=duration, start=self._stop)



class MonoTimer(Timer):
    """
    Class to manage real elaspsed time using time module but with monotonically
    increating time guarantee in spite of system time being retrograded.

    If the system clock is retrograded (moved back in time) while the timer is
    running then time.time() could move to before the start time.
    MonoTimer detects this retrograde and if retro is True then
    retrogrades the start and stop times back Otherwise it raises a TimerRetroError.
    MonoTimer is not able to detect a prograded clock (moved forward in time)

    Attributes:
        ._start is start time in seconds
        ._stop  is stop time in seconds
        ._last is last measured time in seconds with retrograde handling
        .retro is boolean If True retrograde ._start and ._stop when time is retrograded.

    Properties:
        .duration is float time duration in seconds of timer from ._start to ._stop
        .elaspsed is float time elasped in seconds since ._start
        .remaining is float time remaining in seconds until ._stop
        .expired is boolean True if expired, False otherwise, i.e. time >= ._stop
        .latest is float latest measured time in seconds with retrograte handling

    methods:
        .start() = start timer at current time returns start time
        .restart() = restart timer at last ._stop so no time lost, returns start time
    """

    def __init__(self, duration=0.0, start=None, retro=True):
        """
        Initialization method for instance.
        Parameters:
            duration in seconds (fractional)
            start is float optional start time in seconds allows starting before
               or after current time
            retro is boolean IF True automaticall shift timer whenever
                retrograded clock detected Otherwise ignore
        """
        self._start = float(start) if start is not None else time.time()
        self._stop = self._start + float(duration)  # need for default duration
        self._last = self._start
        self.retro = True if retro else False  #  ensure boolean
        self.start(duration=duration, start=start)


    @property
    def elapsed(self):
        """
        elapsed time property getter,
        Returns elapsed time in seconds (fractional) since ._start.
        """
        return (self.latest - self._start)


    @property
    def remaining(self):
        """
        remaining time property getter,
        Returns remaining time in seconds (fractional) before ._stop.
        """
        return (self._stop - self.latest)


    @property
    def expired(self):
        """
        Returns True if timer has expired, False otherwise.
        .latest >= ._stop,
        """
        return (self.latest >= self._stop)

    @property
    def latest(self):
        """
        latest measured time property getter,
        Returns latest measured time in seconds adjusted for retrograded system time.
        """
        delta = time.time() - self._last  # current time - last time checked
        if delta < 0:  # system clock has retrograded
            if self.retro:
                self._start += delta
                self._stop += delta
            else:
                raise RetroTimerError("System time retrograded by {0} seconds"
                                      " while timer running.".format(delta))

        self._last += delta
        return self._last





# -*- encoding: utf-8 -*-
"""
hio.core.packeting Module


"""


import struct
from binascii import hexlify
from collections import deque, namedtuple
import enum

from .. import hioing

#from ...aid.byting import packify, packifyInto, unpackify
# see hio versions in help.packing



class Part(hioing.MixIn):
    """
    Base class for packet part classes with .packed .size and len
    """
    Size = 0

    def __init__(self, size=None, packed=None, **kwa):
        """
        Initialization method for instance.

        Parameters:
            size is initial size of .packed if packed not provided
            packed is initial .packed

        Attributes:
            .packed is bytearray of packed binary data

        Properties:
            .size is length of .packed

        """
        super(Part, self).__init__(**kwa)

        if packed is None:
            size = size if size is not None else self.Size
            self.packed = bytearray([0 for i in range(0, size)])
        else:
            self.packed = bytearray(packed)

    def __len__(self):
        """
        Returns the length of .packed
        """
        return len(self.packed)

    @property
    def size(self):
        """
        Property size
        """
        return self.__len__()

    def show(self):
        """
        Returns descriptive string for display purposes
        """
        name = self.__class__.__name__
        result = ("    {0}: packed=0x{1}\n".format(name,
                                        hexlify(self.packed).decode('ascii')))
        return result


class PackerPart(Part):
    """
    Base class for packet packer part classes with .packer and .fmt
    .fmt is the struct string format for the fixed size portion of part
    """
    Format = '!'  # default empty packer struct format string for packed

    def __init__(self, fmt=None, raw=None, **kwa):
        """
        Initialization method for instance.

        Inherited Parameters:
            size is initial size of .packed

        Parameters:
            fmt is struct format string to pack into .packed
            raw is input bytearray of data to parse(unpack)

        Inherited Attributes:
            .packed is bytearray of packed binary data using .packer

        Attributes:
            .fmt is struct format string
            .packer is compiled struct packer

        Class Attributes:
            .Format is struct packer format string for packed

        Inherited Properties:
            .size is length of .packed

        """
        self.fmt = fmt if fmt is not None else self.Format
        self.packer = struct.Struct(self.fmt) #precompile struct for packing unpacking
        kwa['size'] = self.packer.size  # override size to match .packer.size
        super(PackerPart, self).__init__(**kwa)

        if raw is not None:
            self.parse(raw=raw)

    def verifySize(self, raw=bytearray(b'')):
        """
        Return True if len(raw) is at least long enough for packed size
        """
        return (len(raw) >= (self.packer.size))

    def parse(self, raw):
        """Parse raw bytearray and assign to fields
           Return offset into raw of unparsed portion
           Base method to be overridden in subclass
        """
        if (raw is None) or (not self.verifySize(raw)):
            raise ValueError("Parse Packer: Not enough raw data for packer. "
                             "Need {0} bytes, got {1} bytes.".format(self.size,
                                                                     len(raw)))

        result = self.packer.unpack_from(raw)  # result should be empty tuple
        self.packed[:] = raw[0:self.packer.size]

        return self.size #return offset to start of unparsed portion of data

    def pack(self, **kwa):
        """
        Return .packed with data if any
        Base method to be overridden in sub class
        """

        self.packer.pack_into(self.packed, 0)  # empty

        if self.size != self.packer.size :
            raise ValueError("Build Packer: size packed={0} not match "
                             "format={1}".format(self.size, self.packer.size))
        return self.packed


class PackifierPart(Part):
    """
    Base class for packet packifier part classes with packify .fmt
    .fmt is the packify string format for the fixed size portion of part
    packify allows bit field packing

    packify/unpackify format is string of white space separated bit field lengths
    The packed values are provided as sequence of bit field values
    that are packed into bytearray of size bytes using fmt string.

    Each white space separated field of fmt is the length of the associated bit field
    If not provided size is the least integer number of bytes that hold the fmt.
    If reverse is true reverse the order of the bytes in the byte array before
    returning. This is useful for converting between bigendian and littleendian.

    Assumes unsigned fields values.
    Assumes network big endian so first fields element is high order bits.
    Each field in format string is number of bits for the associated bit field
    Fields with length of 1 are treated as has having boolean truthy field values
       that is,   nonzero is True and packs as a 1
    for 2+ length bit fields the field element is truncated to the number of
       low order bits in the bit field
    if sum of number of bits in fmt less than size bytes then the last byte in
       the bytearray is right zero padded
    if sum of number of bits in fmt greater than size bytes returns exception
    to pad just use 0 value in source field.
    example
    packify("1 3 2 2", (True, 4, 0, 3)). returns bytearry([0xc3])
    """
    Format = ''  # default packer struct format string for packed

    def __init__(self, fmt=None, raw=None, **kwa):
        """
        Initialization method for instance.

        Inherited Parameters:
            size is initial size of .packed

        Parameters:
            fmt is packify format string to pack into .packed
            raw is input bytearray of data to parse(unpack)

        Inherited Attributes:
            .packed is bytearray of packed binary data

        Attributes:
            .fmt is packify format string
            .fmtSize is size given by format string

        Inherited Properties:
            .size is length of .packed

        Properties
            .fmtSize is size given by .fmt

        """
        self.fmt = fmt if fmt is not None else self.Format
        kwa['size'] = self.fmtSize  # override size to match packify size of whole bytes
        super(PackifierPart, self).__init__(**kwa)

        if raw is not None:
            self.parse(raw=raw)

    @property
    def fmtSize(self):
        """
        Property fmtSize
        """
        tbfl = sum((int(x) for x in self.fmt.split()))
        size = (tbfl // 8) + 1 if tbfl % 8 else tbfl // 8
        return size

    def verifySize(self, raw=bytearray(b'')):
        """
        Return True if len(raw) is at least long enough for formatted size
        """
        return (len(raw) >= (self.fmtSize))

    def parse(self, raw):
        """Parse raw bytearray and assign to fields
           Return offset into raw of unparsed portion
           Base method to be overridden in subclass
        """
        if (not raw) or (not self.verifySize(raw)):
            raise ValueError("Parse Packifier: Not enough raw data for packifier. "
                             "Need {0} bytes, got {1} bytes.".format(self.size,
                                                                     len(raw)))

        result = unpackify(self.fmt, raw, boolean=True, size=self.fmtSize)  # empty result
        self.packed[:] = raw[0:self.size]

        return self.size #return offset to start of unparsed portion of data

    def pack(self, **kwa):
        """
        Return .packed with data if any
        Base method to be overridden in sub class
        """
        size = packifyInto(self.packed, fmt=self.fmt, fields=())

        if self.size != size :
            raise ValueError("Build Packifier: size packed={0} not match "
                             "format={1}".format(self.size, size))
        return self.packed

    def show(self):
        """
        Returns descriptive string for display purposes
        """
        name = self.__class__.__name__
        result = ("    {0}: packed=0x{1}\n".format(name,
                                        hexlify(self.packed).decode('ascii')))
        return result


class PacketPart(Part):
    """
    PacketPart base class for parts of packets.
    Allows PacketPart to reference other parts of its Packet
    """

    def __init__(self, packet=None, **kwa):
        """
        Initialization method for instance.
        Base class method to be overridden in subclass
        Need to add parts to packet in subclass

        Inherited Parameters:
            size is initial size of .packed

        Parameters:
            packet is Packet instance that holds this part

        Inherited Attributes:
            .packed is bytearray of packed binary data

        Attributes:
            .packet is Packet instance that holds this part

        Properties:
            .size is length of .packed

        """
        self.packet = packet  # do this first  in case mixin needs attribute
        super(PacketPart, self).__init__(**kwa)


    def show(self):
        """
        Returns descriptive string for display purposes
        """
        name = self.__class__.__name__
        result = ("    {0}: packed=0x{1}\n".format(name,
                                        hexlify(self.packed).decode('ascii')))
        return result


class Packet(Part):
    """
    Packet base class
    Allows packet to reference its stack
    """

    def __init__(self, stack=None, **kwa):
        """
        Initialization method for instance.
        Base class method to be overridden in subclass
        Need to add parts to packet in subclass

        Inherited Parameters:
            size is initial size of .packed

        Parameters:
            stack is I/O stack that handles this packet

        Inherited Attributes:
            .packed is bytearray of packed binary data

        Attributes:
            .stack is I/O stack that handles this packet

        Inherited Properties:
            .size is length of .packed


        """
        super(Packet, self).__init__(**kwa)
        self.stack = stack


    def parse(self, raw):
        """
        Parse raw data into .packed
        """
        self.packed = bytearray(raw)
        return self.size

    def pack(self):
        """
        Pack into .packed
        """
        return self.packed




# -*- encoding: utf-8 -*-
"""
hio.core.tcp Package
"""

from .clienting import openClient, Client, ClientTls, ClientDoer
from .serving import openServer, Server, ServerTls, Remoter, ServerDoer, EchoServerDoer
